#!/bin/sh
#
# build module, install, and run tests:
#   - expecting $NEST_INSTALL_DIR to be set
#   - will create build directory ~/build/nest-MODNAME-NOW/
#   - will install in $NEST_INSTALL_DIR
#

modname="unsafemodule"
now=$(date "+%F-%H-%M")

set -xe

# ----------------------------------------------------------------------
# check environment

if [ -z "$NEST_INSTALL_DIR"] ; then
  set +xe
  echo "$0: error: \$NEST_INSTALL_DIR not set"
  exit 1
fi

# ----------------------------------------------------------------------
# setup directories

cd $(dirname $0)
code_dir=$(pwd)
code_dir_name=$(basename ${code_dir})
build_dir=/calc/$(whoami)/build
build_base="nest-${modname}-${now}"

[ ! -d ${build_dir} ] && mkdir ${build_dir}
cd ${build_dir}

# create build directory, fail if it exists (may not start build more often
# than once per second)

if [ -d ${build_base} ] ; then
  set +xe
  echo "build directory exists, aborting."
  exit -1
fi

mkdir ${build_base}
cd ${build_base}

# copy code directory, create build directory
cp -R ${code_dir} .
build="build"
mkdir $build
cd $build

# ----------------------------------------------------------------------
# configure

cmake -Dwith-nest=${NEST_INSTALL_DIR}/bin/nest-config -Dwith-python=3 ../${code_dir_name}

if [ $? -ne 0 ]; then
  set +xe
  echo "cmake failed, aborting."
  exit -1
fi

# ----------------------------------------------------------------------
# make
make -j $(nproc)

if [ $? -ne 0 ]; then
  set +xe
  echo "make failed, aborting."
  exit -1
fi

# ----------------------------------------------------------------------
# install

make install

if [ $? -ne 0 ]; then
  set +xe
  echo "make install failed, aborting."
  exit -1
fi

# ----------------------------------------------------------------------
# test

cd ${code_dir}
for f in tests/*.py ; do
    python3 $f

    if [ $? -ne 0 ]; then
      set +xe
      echo "test $f failed, aborting."
      exit -1
    fi
done

# ----------------------------------------------------------------------
# print how to use

set +xe

echo ""
echo "----------------------------------------------------------------------"
echo "installation and tests ran successfully."
echo ""

echo "to import module, use:"
echo ""
echo "    nest.Install(\"${modname}\")"
echo ""
