/*
 *  unsafemodule.cpp
 */

#include "unsafemodule.h"

// Generated includes:
#include "config.h"

// include headers with your own stuff
#include "eif_psc_exp.h"
#include "aeif_psc_delta_unsafe.h"
#include "aeif_psc_exp_unsafe.h"
#include "aeif_psc_alpha_unsafe.h"
#include "aeif_cond_exp_unsafe.h"
#include "aeif_cond_alpha_unsafe.h"

// Includes from nestkernel:
#include "connection_manager_impl.h"
#include "connector_model_impl.h"
#include "dynamicloader.h"
#include "exceptions.h"
#include "genericmodel.h"
#include "genericmodel_impl.h"
#include "kernel_manager.h"
#include "model.h"
#include "model_manager_impl.h"
#include "nestmodule.h"
#include "target_identifier.h"

// Includes from sli:
#include "booldatum.h"
#include "integerdatum.h"
#include "sliexceptions.h"
#include "tokenarray.h"

// -- Interface to dynamic module loader ---------------------------------------

/*
 * There are three scenarios, in which UnsafeModule can be loaded by NEST:
 *
 * 1) When loading your module with `Install`, the dynamic module loader must
 * be able to find your module. You make the module known to the loader by
 * defining an instance of your module class in global scope. (LTX_MODULE is
 * defined) This instance must have the name
 *
 * <modulename>_LTX_mod
 *
 * The dynamicloader can then load modulename and search for symbol "mod" in it.
 *
 * 2) When you link the library dynamically with NEST during compilation, a new
 * object has to be created. In the constructor the DynamicLoaderModule will
 * register your module. (LINKED_MODULE is defined)
 *
 * 3) When you link the library statically with NEST during compilation, the
 * registration will take place in the file `static_modules.h`, which is
 * generated by cmake.
 */
#if defined( LTX_MODULE ) | defined( LINKED_MODULE )
unsafemodule_ns::UnsafeModule unsafemodule_LTX_mod;
#endif
// -- DynModule functions ------------------------------------------------------

unsafemodule_ns::UnsafeModule::UnsafeModule()
{
#ifdef LINKED_MODULE
  // register this module at the dynamic loader
  // this is needed to allow for linking in this module at compile time
  // all registered modules will be initialized by the main app's dynamic loader
  nest::DynamicLoaderModule::registerLinkedModule( this );
#endif
}

unsafemodule_ns::UnsafeModule::~UnsafeModule()
{
}

const std::string
unsafemodule_ns::UnsafeModule::name( void ) const
{
  return std::string( "Unsafe Module" ); // Return name of the module
}

const std::string
unsafemodule_ns::UnsafeModule::commandstring( void ) const
{
  // Instruct the interpreter to load unsafemodule-init.sli
  return std::string( "(unsafemodule-init) run" );
}

//-------------------------------------------------------------------------------------

void
unsafemodule_ns::UnsafeModule::init( SLIInterpreter* i )
{
  // register neuron models
  nest::kernel().model_manager.register_node_model<nest::eif_psc_exp>("eif_psc_exp");
  nest::kernel().model_manager.register_node_model<nest::aeif_psc_delta_unsafe>("aeif_psc_delta_unsafe");
  nest::kernel().model_manager.register_node_model<nest::aeif_psc_delta_unsafe>("aeif_psc_exp_unsafe");
  nest::kernel().model_manager.register_node_model<nest::aeif_psc_delta_unsafe>("aeif_psc_alpha_unsafe");
  nest::kernel().model_manager.register_node_model<nest::aeif_psc_delta_unsafe>("aeif_cond_exp_unsafe");
  nest::kernel().model_manager.register_node_model<nest::aeif_psc_delta_unsafe>("aeif_cond_alpha_unsafe");

  // register synapse models

} // UnsafeModule::init()
