#!/usr/bin/env python3

from cycler import cycler
from matplotlib import cm
from matplotlib.collections import LineCollection
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
from mpl_toolkits.axes_grid1.inset_locator import inset_axes
import numpy as np


def __gen_cc(color_list, num_c, num_ls, is_cyclic=False, rev=False):
    N = len(color_list)
    assert 0 < num_c <= N

    if num_c < N:
        if not is_cyclic:
            ind = np.linspace(0, N-1, num=num_c, dtype=int)
        else:
            ind = np.linspace(0, N-1, num=num_c+1, dtype=int)[:-1]

        colors = [color_list[i] for i in ind]

    else:
        colors = color_list

    if rev:
        colors = colors[::-1]

    cc = cycler(color=colors)

    if num_ls > 1:
        cc = cc * cycler(linestyle=__linestyles[:num_ls])

    return cc


def __gen_cc_4_ls(color_list, num_c, is_cyclic=False, rev=False):
    cc_x1 = __gen_cc(color_list, num_c, num_ls=1, is_cyclic=is_cyclic, rev=rev)
    cc_x2 = __gen_cc(color_list, num_c, num_ls=2, is_cyclic=is_cyclic, rev=rev)
    cc_x3 = __gen_cc(color_list, num_c, num_ls=3, is_cyclic=is_cyclic, rev=rev)
    cc_x4 = __gen_cc(color_list, num_c, num_ls=4, is_cyclic=is_cyclic, rev=rev)
    return cc_x1, cc_x1, cc_x2, cc_x3, cc_x4


def __get_colors_as_list(cmap):
    return [cmap(n) for n in range(cmap.N)]


__cividis_colors = __get_colors_as_list(cm.cividis)
__inferno_colors = __get_colors_as_list(cm.inferno)
__magma_colors = __get_colors_as_list(cm.magma)
__plasma_colors = __get_colors_as_list(cm.plasma)
__terrain_colors = __get_colors_as_list(cm.terrain)
__viridis_colors = __get_colors_as_list(cm.viridis)

__xkcd_colors = [
        'xkcd:azure',
        'xkcd:goldenrod',
        'xkcd:green',
        'xkcd:salmon',
        #'xkcd:indigo',
        'xkcd:darkblue',
        'xkcd:grey',
        'xkcd:sienna',
        'xkcd:orchid',
]

__linestyles = [
        '-',
        '--',
        ':',
        '-.',
]

cc_cividis4, cc_cividis4x1, cc_cividis4x2, cc_cividis4x3, cc_cividis4x4 = __gen_cc_4_ls(__cividis_colors, 4, is_cyclic=True)
cc_cividis6, cc_cividis6x1, cc_cividis6x2, cc_cividis6x3, cc_cividis6x4 = __gen_cc_4_ls(__cividis_colors, 6, is_cyclic=True)
cc_cividis8, cc_cividis8x1, cc_cividis8x2, cc_cividis8x3, cc_cividis8x4 = __gen_cc_4_ls(__cividis_colors, 8, is_cyclic=True)
cc_cividis10, cc_cividis10x1, cc_cividis10x2, cc_cividis10x3, cc_cividis10x4 = __gen_cc_4_ls(__cividis_colors, 10, is_cyclic=True)
cc_cividis12, cc_cividis12x1, cc_cividis12x2, cc_cividis12x3, cc_cividis12x4 = __gen_cc_4_ls(__cividis_colors, 12, is_cyclic=True)
cc_cividis16, cc_cividis16x1, cc_cividis16x2, cc_cividis16x3, cc_cividis16x4 = __gen_cc_4_ls(__cividis_colors, 16, is_cyclic=True)
cc_cividis32, cc_cividis32x1, cc_cividis32x2, cc_cividis32x3, cc_cividis32x4 = __gen_cc_4_ls(__cividis_colors, 32, is_cyclic=True)

cc_rcividis4, cc_rcividis4x1, cc_rcividis4x2, cc_rcividis4x3, cc_rcividis4x4 = __gen_cc_4_ls(__cividis_colors, 4, is_cyclic=True, rev=True)
cc_rcividis6, cc_rcividis6x1, cc_rcividis6x2, cc_rcividis6x3, cc_rcividis6x4 = __gen_cc_4_ls(__cividis_colors, 6, is_cyclic=True, rev=True)
cc_rcividis8, cc_rcividis8x1, cc_rcividis8x2, cc_rcividis8x3, cc_rcividis8x4 = __gen_cc_4_ls(__cividis_colors, 8, is_cyclic=True, rev=True)
cc_rcividis10, cc_rcividis10x1, cc_rcividis10x2, cc_rcividis10x3, cc_rcividis10x4 = __gen_cc_4_ls(__cividis_colors, 10, is_cyclic=True, rev=True)
cc_rcividis12, cc_rcividis12x1, cc_rcividis12x2, cc_rcividis12x3, cc_rcividis12x4 = __gen_cc_4_ls(__cividis_colors, 12, is_cyclic=True, rev=True)
cc_rcividis16, cc_rcividis16x1, cc_rcividis16x2, cc_rcividis16x3, cc_rcividis16x4 = __gen_cc_4_ls(__cividis_colors, 16, is_cyclic=True, rev=True)
cc_rcividis32, cc_rcividis32x1, cc_rcividis32x2, cc_rcividis32x3, cc_rcividis32x4 = __gen_cc_4_ls(__cividis_colors, 32, is_cyclic=True, rev=True)

cc_inferno4, cc_inferno4x1, cc_inferno4x2, cc_inferno4x3, cc_inferno4x4 = __gen_cc_4_ls(__inferno_colors, 4, is_cyclic=True)
cc_inferno6, cc_inferno6x1, cc_inferno6x2, cc_inferno6x3, cc_inferno6x4 = __gen_cc_4_ls(__inferno_colors, 6, is_cyclic=True)
cc_inferno8, cc_inferno8x1, cc_inferno8x2, cc_inferno8x3, cc_inferno8x4 = __gen_cc_4_ls(__inferno_colors, 8, is_cyclic=True)
cc_inferno10, cc_inferno10x1, cc_inferno10x2, cc_inferno10x3, cc_inferno10x4 = __gen_cc_4_ls(__inferno_colors, 10, is_cyclic=True)
cc_inferno12, cc_inferno12x1, cc_inferno12x2, cc_inferno12x3, cc_inferno12x4 = __gen_cc_4_ls(__inferno_colors, 12, is_cyclic=True)
cc_inferno16, cc_inferno16x1, cc_inferno16x2, cc_inferno16x3, cc_inferno16x4 = __gen_cc_4_ls(__inferno_colors, 16, is_cyclic=True)
cc_inferno32, cc_inferno32x1, cc_inferno32x2, cc_inferno32x3, cc_inferno32x4 = __gen_cc_4_ls(__inferno_colors, 32, is_cyclic=True)

cc_rinferno4, cc_rinferno4x1, cc_rinferno4x2, cc_rinferno4x3, cc_rinferno4x4 = __gen_cc_4_ls(__inferno_colors, 4, is_cyclic=True, rev=True)
cc_rinferno6, cc_rinferno6x1, cc_rinferno6x2, cc_rinferno6x3, cc_rinferno6x4 = __gen_cc_4_ls(__inferno_colors, 6, is_cyclic=True, rev=True)
cc_rinferno8, cc_rinferno8x1, cc_rinferno8x2, cc_rinferno8x3, cc_rinferno8x4 = __gen_cc_4_ls(__inferno_colors, 8, is_cyclic=True, rev=True)
cc_rinferno10, cc_rinferno10x1, cc_rinferno10x2, cc_rinferno10x3, cc_rinferno10x4 = __gen_cc_4_ls(__inferno_colors, 10, is_cyclic=True, rev=True)
cc_rinferno12, cc_rinferno12x1, cc_rinferno12x2, cc_rinferno12x3, cc_rinferno12x4 = __gen_cc_4_ls(__inferno_colors, 12, is_cyclic=True, rev=True)
cc_rinferno16, cc_rinferno16x1, cc_rinferno16x2, cc_rinferno16x3, cc_rinferno16x4 = __gen_cc_4_ls(__inferno_colors, 16, is_cyclic=True, rev=True)
cc_rinferno32, cc_rinferno32x1, cc_rinferno32x2, cc_rinferno32x3, cc_rinferno32x4 = __gen_cc_4_ls(__inferno_colors, 32, is_cyclic=True, rev=True)

cc_magma4, cc_magma4x1, cc_magma4x2, cc_magma4x3, cc_magma4x4 = __gen_cc_4_ls(__magma_colors, 4, is_cyclic=True)
cc_magma6, cc_magma6x1, cc_magma6x2, cc_magma6x3, cc_magma6x4 = __gen_cc_4_ls(__magma_colors, 6, is_cyclic=True)
cc_magma8, cc_magma8x1, cc_magma8x2, cc_magma8x3, cc_magma8x4 = __gen_cc_4_ls(__magma_colors, 8, is_cyclic=True)
cc_magma10, cc_magma10x1, cc_magma10x2, cc_magma10x3, cc_magma10x4 = __gen_cc_4_ls(__magma_colors, 10, is_cyclic=True)
cc_magma12, cc_magma12x1, cc_magma12x2, cc_magma12x3, cc_magma12x4 = __gen_cc_4_ls(__magma_colors, 12, is_cyclic=True)
cc_magma16, cc_magma16x1, cc_magma16x2, cc_magma16x3, cc_magma16x4 = __gen_cc_4_ls(__magma_colors, 16, is_cyclic=True)
cc_magma32, cc_magma32x1, cc_magma32x2, cc_magma32x3, cc_magma32x4 = __gen_cc_4_ls(__magma_colors, 32, is_cyclic=True)

cc_rmagma4, cc_rmagma4x1, cc_rmagma4x2, cc_rmagma4x3, cc_rmagma4x4 = __gen_cc_4_ls(__magma_colors, 4, is_cyclic=True, rev=True)
cc_rmagma6, cc_rmagma6x1, cc_rmagma6x2, cc_rmagma6x3, cc_rmagma6x4 = __gen_cc_4_ls(__magma_colors, 6, is_cyclic=True, rev=True)
cc_rmagma8, cc_rmagma8x1, cc_rmagma8x2, cc_rmagma8x3, cc_rmagma8x4 = __gen_cc_4_ls(__magma_colors, 8, is_cyclic=True, rev=True)
cc_rmagma10, cc_rmagma10x1, cc_rmagma10x2, cc_rmagma10x3, cc_rmagma10x4 = __gen_cc_4_ls(__magma_colors, 10, is_cyclic=True, rev=True)
cc_rmagma12, cc_rmagma12x1, cc_rmagma12x2, cc_rmagma12x3, cc_rmagma12x4 = __gen_cc_4_ls(__magma_colors, 12, is_cyclic=True, rev=True)
cc_rmagma16, cc_rmagma16x1, cc_rmagma16x2, cc_rmagma16x3, cc_rmagma16x4 = __gen_cc_4_ls(__magma_colors, 16, is_cyclic=True, rev=True)
cc_rmagma32, cc_rmagma32x1, cc_rmagma32x2, cc_rmagma32x3, cc_rmagma32x4 = __gen_cc_4_ls(__magma_colors, 32, is_cyclic=True, rev=True)

cc_plasma4, cc_plasma4x1, cc_plasma4x2, cc_plasma4x3, cc_plasma4x4 = __gen_cc_4_ls(__plasma_colors, 4, is_cyclic=True)
cc_plasma6, cc_plasma6x1, cc_plasma6x2, cc_plasma6x3, cc_plasma6x4 = __gen_cc_4_ls(__plasma_colors, 6, is_cyclic=True)
cc_plasma8, cc_plasma8x1, cc_plasma8x2, cc_plasma8x3, cc_plasma8x4 = __gen_cc_4_ls(__plasma_colors, 8, is_cyclic=True)
cc_plasma10, cc_plasma10x1, cc_plasma10x2, cc_plasma10x3, cc_plasma10x4 = __gen_cc_4_ls(__plasma_colors, 10, is_cyclic=True)
cc_plasma12, cc_plasma12x1, cc_plasma12x2, cc_plasma12x3, cc_plasma12x4 = __gen_cc_4_ls(__plasma_colors, 12, is_cyclic=True)
cc_plasma16, cc_plasma16x1, cc_plasma16x2, cc_plasma16x3, cc_plasma16x4 = __gen_cc_4_ls(__plasma_colors, 16, is_cyclic=True)
cc_plasma32, cc_plasma32x1, cc_plasma32x2, cc_plasma32x3, cc_plasma32x4 = __gen_cc_4_ls(__plasma_colors, 32, is_cyclic=True)

cc_rplasma4, cc_rplasma4x1, cc_rplasma4x2, cc_rplasma4x3, cc_rplasma4x4 = __gen_cc_4_ls(__plasma_colors, 4, is_cyclic=True, rev=True)
cc_rplasma6, cc_rplasma6x1, cc_rplasma6x2, cc_rplasma6x3, cc_rplasma6x4 = __gen_cc_4_ls(__plasma_colors, 6, is_cyclic=True, rev=True)
cc_rplasma8, cc_rplasma8x1, cc_rplasma8x2, cc_rplasma8x3, cc_rplasma8x4 = __gen_cc_4_ls(__plasma_colors, 8, is_cyclic=True, rev=True)
cc_rplasma10, cc_rplasma10x1, cc_rplasma10x2, cc_rplasma10x3, cc_rplasma10x4 = __gen_cc_4_ls(__plasma_colors, 10, is_cyclic=True, rev=True)
cc_rplasma12, cc_rplasma12x1, cc_rplasma12x2, cc_rplasma12x3, cc_rplasma12x4 = __gen_cc_4_ls(__plasma_colors, 12, is_cyclic=True, rev=True)
cc_rplasma16, cc_rplasma16x1, cc_rplasma16x2, cc_rplasma16x3, cc_rplasma16x4 = __gen_cc_4_ls(__plasma_colors, 16, is_cyclic=True, rev=True)
cc_rplasma32, cc_rplasma32x1, cc_rplasma32x2, cc_rplasma32x3, cc_rplasma32x4 = __gen_cc_4_ls(__plasma_colors, 32, is_cyclic=True, rev=True)

cc_terrain4, cc_terrain4x1, cc_terrain4x2, cc_terrain4x3, cc_terrain4x4 = __gen_cc_4_ls(__terrain_colors, 4, is_cyclic=True)
cc_terrain6, cc_terrain6x1, cc_terrain6x2, cc_terrain6x3, cc_terrain6x4 = __gen_cc_4_ls(__terrain_colors, 6, is_cyclic=True)
cc_terrain8, cc_terrain8x1, cc_terrain8x2, cc_terrain8x3, cc_terrain8x4 = __gen_cc_4_ls(__terrain_colors, 8, is_cyclic=True)
cc_terrain10, cc_terrain10x1, cc_terrain10x2, cc_terrain10x3, cc_terrain10x4 = __gen_cc_4_ls(__terrain_colors, 10, is_cyclic=True)
cc_terrain12, cc_terrain12x1, cc_terrain12x2, cc_terrain12x3, cc_terrain12x4 = __gen_cc_4_ls(__terrain_colors, 12, is_cyclic=True)
cc_terrain16, cc_terrain16x1, cc_terrain16x2, cc_terrain16x3, cc_terrain16x4 = __gen_cc_4_ls(__terrain_colors, 16, is_cyclic=True)
cc_terrain32, cc_terrain32x1, cc_terrain32x2, cc_terrain32x3, cc_terrain32x4 = __gen_cc_4_ls(__terrain_colors, 32, is_cyclic=True)

cc_rterrain4, cc_rterrain4x1, cc_rterrain4x2, cc_rterrain4x3, cc_rterrain4x4 = __gen_cc_4_ls(__terrain_colors, 4, is_cyclic=True, rev=True)
cc_rterrain6, cc_rterrain6x1, cc_rterrain6x2, cc_rterrain6x3, cc_rterrain6x4 = __gen_cc_4_ls(__terrain_colors, 6, is_cyclic=True, rev=True)
cc_rterrain8, cc_rterrain8x1, cc_rterrain8x2, cc_rterrain8x3, cc_rterrain8x4 = __gen_cc_4_ls(__terrain_colors, 8, is_cyclic=True, rev=True)
cc_rterrain10, cc_rterrain10x1, cc_rterrain10x2, cc_rterrain10x3, cc_rterrain10x4 = __gen_cc_4_ls(__terrain_colors, 10, is_cyclic=True, rev=True)
cc_rterrain12, cc_rterrain12x1, cc_rterrain12x2, cc_rterrain12x3, cc_rterrain12x4 = __gen_cc_4_ls(__terrain_colors, 12, is_cyclic=True, rev=True)
cc_rterrain16, cc_rterrain16x1, cc_rterrain16x2, cc_rterrain16x3, cc_rterrain16x4 = __gen_cc_4_ls(__terrain_colors, 16, is_cyclic=True, rev=True)
cc_rterrain32, cc_rterrain32x1, cc_rterrain32x2, cc_rterrain32x3, cc_rterrain32x4 = __gen_cc_4_ls(__terrain_colors, 32, is_cyclic=True, rev=True)

cc_viridis4, cc_viridis4x1, cc_viridis4x2, cc_viridis4x3, cc_viridis4x4 = __gen_cc_4_ls(__viridis_colors, 4, is_cyclic=True)
cc_viridis6, cc_viridis6x1, cc_viridis6x2, cc_viridis6x3, cc_viridis6x4 = __gen_cc_4_ls(__viridis_colors, 6, is_cyclic=True)
cc_viridis8, cc_viridis8x1, cc_viridis8x2, cc_viridis8x3, cc_viridis8x4 = __gen_cc_4_ls(__viridis_colors, 8, is_cyclic=True)
cc_viridis10, cc_viridis10x1, cc_viridis10x2, cc_viridis10x3, cc_viridis10x4 = __gen_cc_4_ls(__viridis_colors, 10, is_cyclic=True)
cc_viridis12, cc_viridis12x1, cc_viridis12x2, cc_viridis12x3, cc_viridis12x4 = __gen_cc_4_ls(__viridis_colors, 12, is_cyclic=True)
cc_viridis16, cc_viridis16x1, cc_viridis16x2, cc_viridis16x3, cc_viridis16x4 = __gen_cc_4_ls(__viridis_colors, 16, is_cyclic=True)
cc_viridis32, cc_viridis32x1, cc_viridis32x2, cc_viridis32x3, cc_viridis32x4 = __gen_cc_4_ls(__viridis_colors, 32, is_cyclic=True)

cc_rviridis4, cc_rviridis4x1, cc_rviridis4x2, cc_rviridis4x3, cc_rviridis4x4 = __gen_cc_4_ls(__viridis_colors, 4, is_cyclic=True, rev=True)
cc_rviridis6, cc_rviridis6x1, cc_rviridis6x2, cc_rviridis6x3, cc_rviridis6x4 = __gen_cc_4_ls(__viridis_colors, 6, is_cyclic=True, rev=True)
cc_rviridis8, cc_rviridis8x1, cc_rviridis8x2, cc_rviridis8x3, cc_rviridis8x4 = __gen_cc_4_ls(__viridis_colors, 8, is_cyclic=True, rev=True)
cc_rviridis10, cc_rviridis10x1, cc_rviridis10x2, cc_rviridis10x3, cc_rviridis10x4 = __gen_cc_4_ls(__viridis_colors, 10, is_cyclic=True, rev=True)
cc_rviridis12, cc_rviridis12x1, cc_rviridis12x2, cc_rviridis12x3, cc_rviridis12x4 = __gen_cc_4_ls(__viridis_colors, 12, is_cyclic=True, rev=True)
cc_rviridis16, cc_rviridis16x1, cc_rviridis16x2, cc_rviridis16x3, cc_rviridis16x4 = __gen_cc_4_ls(__viridis_colors, 16, is_cyclic=True, rev=True)
cc_rviridis32, cc_rviridis32x1, cc_rviridis32x2, cc_rviridis32x3, cc_rviridis32x4 = __gen_cc_4_ls(__viridis_colors, 32, is_cyclic=True, rev=True)

cc_xkcd4, cc_xkcd4x1, cc_xkcd4x2, cc_xkcd4x3, cc_xkcd4x4 = __gen_cc_4_ls(__xkcd_colors[:4], 4)
cc_xkcd6, cc_xkcd6x1, cc_xkcd6x2, cc_xkcd6x3, cc_xkcd6x4 = __gen_cc_4_ls(__xkcd_colors[:6], 6)
cc_xkcd8, cc_xkcd8x1, cc_xkcd8x2, cc_xkcd8x3, cc_xkcd8x4 = __gen_cc_4_ls(__xkcd_colors[:8], 8)

cc_rxkcd4, cc_rxkcd4x1, cc_rxkcd4x2, cc_rxkcd4x3, cc_rxkcd4x4 = __gen_cc_4_ls(__xkcd_colors[:4], 4, rev=True)
cc_rxkcd6, cc_rxkcd6x1, cc_rxkcd6x2, cc_rxkcd6x3, cc_rxkcd6x4 = __gen_cc_4_ls(__xkcd_colors[:6], 6, rev=True)
cc_rxkcd8, cc_rxkcd8x1, cc_rxkcd8x2, cc_rxkcd8x3, cc_rxkcd8x4 = __gen_cc_4_ls(__xkcd_colors[:8], 8, rev=True)


def raster_plot(times, ids, pop=None, fig=None, ax=None, t_lim=None, xlabel='', ylabel='', xticks=[], yticks=[], lw=.7, c=None):
    fig = plt.gcf() if fig is None else fig
    ax = plt.gca() if ax is None else ax

    if pop is None:
        pop = (np.min(ids), np.max(ids)+1) if len(ids) > 0 else (0, 1)
    elif len(pop) > 2:
        pop = [min(pop), max(pop)+1]

    # discard unnecessary data
    if t_lim:
        m = np.logical_and(times >= t_lim[0], times <= t_lim[1])
        times = times[m]
        ids = ids[m]
    if pop:
        m = np.logical_and(ids >= pop[0], ids < pop[-1])
        times = times[m]
        ids = ids[m]

    # get actual axes height in pixels
    bbox = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    h_px = bbox.height * fig.dpi

    # calculate marker size
    ms_px = h_px / (pop[1] - pop[0] + 1)
    ms_pt = ms_px * 72. / fig.dpi

    ax.scatter(times, ids+.5, marker='|', s=ms_pt**2, linewidth=lw, c=c)

    if t_lim is None:
        ax.autoscale('x', tight=True)
    else:
        assert len(t_lim) == 2
        ax.set_xlim(*t_lim)
    ax.set_ylim(pop[0], pop[1])
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    if xticks is not None:
        if isinstance(xticks, int):
            ax.locator_params(axis='x', nbins=xticks)
        else:
            ax.set_xticks(xticks)
    if yticks is not None:
        if isinstance(yticks, int):
            ax.locator_params(axis='x', nbins=yticks)
        else:
            ax.set_yticks(yticks)


def grouped_barplot(group_mean_, group_std_=None, ax=None, title=None, c=None, xticklabels=None, legend_args=None, **kwargs):
    """
    Create a bar plot with multiple groups of bars.
    """

    if ax is None:
        ax = plt.gca()

    num_groups = len(group_mean_)
    num_points = len(group_mean_[0])

    if group_std_:
        assert len(group_std_) == num_groups

    for group in group_mean_[1:]:
        assert len(group) == num_points

    if group_std_:
        for group in group_std_:
            assert len(group) == num_points

    ind_ = np.arange(num_points)

    bar_width = 1. / (num_groups+1)
    offset_ = [k*bar_width+bar_width/2 for k in range(num_groups)]

    xtick_offset = (1. - bar_width) / 2.
    xtick_ = ind_ + xtick_offset
    xticklabel_ = [str(ind) for ind in ind_] if xticklabels is None else xticklabels

    c_ = ['C{0:d}'.format(k % 10) for k in range(num_groups)] if c is None else c
    t_ = [None]*num_groups if title is None else title

    for k in range(num_groups):
        c, t, offset = c_[k], t_[k], offset_[k]
        group_mean = group_mean_[k]
        group_std = group_std_[k] if group_std_ else None
        ax.bar(ind_+offset, group_mean, bar_width, yerr=group_std, color=c, label=t, **kwargs)

    kwargs = {'loc': 'best'}
    if legend_args is not None:
        kwargs.update(legend_args)

    ax.legend(**kwargs)
    ax.set_xticks(xtick_)
    ax.set_xticklabels(xticklabel_)


def label_axes(label, fig=None, ax=None, offset=(-.04, .02), fs=32, tex=True):
    if fig is None:
        fig = plt.gcf()

    if ax is None:
        ax = plt.gca()

    bbox = ax.get_position()
    left = bbox.x0
    top = bbox.y0 + bbox.height
    pos = (left+offset[0], top+offset[1])
    if tex:
        label = r'\textsf{' + label + '}'
    ax.text(*pos, label, transform=fig.transFigure, clip_on=False, fontsize=fs, ha='right')

def barplot_annotate_brackets(num1, num2, data, center, height, yerr=None, dh=.05, barh=.05, dh_text=None, fs=None, maxasterix=None):
    """
    Annotate barplot with p-values.

    :param num1: number of left bar to put bracket over
    :param num2: number of right bar to put bracket over
    :param data: string to write or number for generating asterixes
    :param center: centers of all bars (like plt.bar() input)
    :param height: heights of all bars (like plt.bar() input)
    :param yerr: yerrs of all bars (like plt.bar() input)
    :param dh: height offset over bar / bar + yerr in axes coordinates (0 to 1)
    :param barh: bar height in axes coordinates (0 to 1)
    :param dh_text: text offset, default: barh
    :param fs: font size
    :param maxasterix: maximum number of asterixes to write (for very small p-values)
    """

    if type(data) is str:
        text = data
    else:
        # * is p < 0.05
        # ** is p < 0.005
        # *** is p < 0.0005
        # etc.
        text = ''
        p = .05

        while data < p:
            text += '*'
            p /= 10.

            if maxasterix and len(text) == maxasterix:
                break

        if len(text) == 0:
            text = 'n. s.'

    lx, ly = center[num1], max(height[num1], max(height[num1:num2]))
    rx, ry = center[num2], max(height[num2], max(height[num1:num2]))

    if yerr:
        ly += yerr[num1]
        ry += yerr[num2]

    ax_y0, ax_y1 = plt.gca().get_ylim()
    dh *= (ax_y1 - ax_y0)
    barh *= (ax_y1 - ax_y0)

    y = max(ly, ry) + dh

    barx = [lx, lx, rx, rx]
    bary = [y, y+barh, y+barh, y]

    plt.plot(barx, bary, c='black')

    if dh_text is None:
        dh_text = barh
    else:
        dh_text *= (ax_y1 - ax_y0)
    mid = ((lx+rx)/2, y+dh_text)

    kwargs = dict(ha='center', va='bottom')
    if fs is not None:
        kwargs['fontsize'] = fs

    plt.text(*mid, text, **kwargs)


def show_spines(ax=None, left=True, top=True, right=True, bottom=True):
    if ax is None:
        ax = plt.gca()

    ax.spines['left'].set_visible(left)
    ax.spines['top'].set_visible(top)
    ax.spines['right'].set_visible(right)
    ax.spines['bottom'].set_visible(bottom)

    if not bottom:
        ax.set_xticks([])
    if not left:
        ax.set_yticks([])


def inset_text(text, vpos='top', hpos='left', ax=None, ha=None, va=None, **kwargs):
    if ax is None:
        ax = plt.gca()

    if type(hpos) is str:
        if hpos == 'left':
            x = .02
            if ha is None:
                ha = 'left'
        elif hpos in ['center', 'middle']:
            x = .5
            if ha is None:
                ha = 'center'
        elif hpos == 'right':
            x = 1.
            if ha is None:
                ha = 'right'
        else:
            raise ValueError('bad value for hpos: '+hpos)
    else:
        x = hpos
        if ha is None:
            ha = 'center'

    if type(vpos) is str:
        if vpos == 'top':
            y = .99
            if va is None:
                va = 'top'
        elif vpos in ['center', 'middle']:
            y = .5
            if va is None:
                va = 'center'
        elif vpos == 'bottom':
            y = .01
            if va is None:
                va = 'bottom'
        else:
            raise ValueError('bad value for vpos: '+vpos)
    else:
        y = vpos
        if va is None:
            va = 'center'

    ax.text(x, y, text, ha=ha, va=va, transform=ax.transAxes, **kwargs)


def split_axes_vertical(split, fig=None, ax=None, dh=.05, pad=None):
    """
    Split given (or current) axes vertically.

    :param split: integer (number of equal size splits) or list of relative sizes (integers)
    :param fig: figure (default: current figure)
    :param ax: axes (default: current axes)
    :param dh: fraction of new axes to use for spacing between new sub-axes (default: 0.05)
    :param pad: optional extra padding: list of fractions to use for padding (left, bottom, right, top)
    :returns: list of new axes
    :
    """

    if pad:
        assert len(pad) == 4
        pad_l, pad_b, pad_r, pad_t = pad
    else:
        pad_l, pad_b, pad_r, pad_t = 0., 0., 0., 0.

    if fig is None:
        fig = plt.gcf()

    if ax is None:
        ax = plt.gca()

    if type(split) is int:
        num_axes = split
        heights = [1] * num_axes
    else:
        assert type(split) in [list, tuple]
        num_axes = len(split)
        heights = split
        for h in heights:
            assert type(h) is int, 'all passed heights must be integers'
            assert h > 0, 'all passed heights must be positive'

    heights = heights[::-1]  # intuitive order is top down
    sum_heights = sum(heights)

    # hide base axes
    ax.set_xticks([])
    ax.set_yticks([])
    show_spines(ax, False, False, False, False)

    # get base axes position

    bbox = ax.get_position()
    x0, y0, x1, y1, w0, h0 = bbox.x0, bbox.y0, bbox.x1, bbox.y1, bbox.width, bbox.height

    # apply padding

    x0 += pad_l*w0
    y0 += pad_b*h0
    x1 -= pad_r*w0
    y1 -= pad_t*h0

    width = x1 - x0
    height = y1 - y0

    inset_offset = height * dh / (num_axes - 1) if num_axes > 1 else 0  # space between axes
    inset_height_total = height * (1. - dh)   # height of all axes together

    ax_list = []

    for k, h in enumerate(heights):
        # height of current axes
        inset_height = inset_height_total * h / sum_heights

        # y-coordinate of current axes
        inset_y = y0 + k*inset_offset + inset_height_total * sum(heights[:k]) / sum_heights

        # add axes
        ax = fig.add_axes([x0, inset_y, width, inset_height])
        ax_list += [ax]

    return ax_list[::-1]


def split_axes_horizontal(split, fig=None, ax=None, dw=.05, pad=None):
    """
    Split given (or current) axes horizontally.

    :param split: integer (number of equal size splits) or list of relative sizes (integers)
    :param fig: figure (default: current figure)
    :param ax: axes (default: current axes)
    :param dw: fraction of new axes to use for spacing between new sub-axes (default: 0.05)
    :param pad: optional extra padding: list of fractions to use for padding (left, bottom, right, top)
    :returns: list of new axes
    :
    """

    if pad:
        assert len(pad) == 4
        pad_l, pad_b, pad_r, pad_t = pad
    else:
        pad_l, pad_b, pad_r, pad_t = 0., 0., 0., 0.

    if fig is None:
        fig = plt.gcf()

    if ax is None:
        ax = plt.gca()

    if type(split) is int:
        num_axes = split
        widths = [1] * num_axes
    else:
        assert type(split) in [list, tuple]
        num_axes = len(split)
        widths = split
        for w in widths:
            assert type(w) is int, 'all passed widths must be integers'
            assert w > 0, 'all passed widths must be positive'

    sum_widths = sum(widths)

    # hide base axes
    ax.set_xticks([])
    ax.set_yticks([])
    show_spines(ax, False, False, False, False)

    # get base axes position

    bbox = ax.get_position()
    x0, y0, x1, y1, w0, h0 = bbox.x0, bbox.y0, bbox.x1, bbox.y1, bbox.width, bbox.height

    # apply padding

    x0 += pad_l*w0
    y0 += pad_b*h0
    x1 -= pad_r*w0
    y1 -= pad_t*h0

    width = x1 - x0
    height = y1 - y0

    inset_offset = width * dw / (num_axes - 1) if num_axes > 1 else 0  # space between axes
    inset_width_total = width * (1. - dw)   # width of all axes together

    ax_list = []

    for k, w in enumerate(widths):
        # width of current axes
        inset_width = inset_width_total * w / sum_widths

        # y-coordinate of current axes
        inset_x = x0 + k*inset_offset + inset_width_total * sum(widths[:k]) / sum_widths

        # add axes
        ax = fig.add_axes([inset_x, y0, inset_width, height])
        ax_list += [ax]

    return ax_list


def inset_new_axes(x0, y0, width, height=None, fig=None, ax=None, ticksize=None):
    # if height is None, axes will have square shape

    if fig is None:
        fig = plt.gcf()

    if ax is None:
        ax = plt.gca()

    # get base axes position

    bbox = ax.get_position()
    bx0, by0, bx1, by1 = bbox.x0, bbox.y0, bbox.x1, bbox.y1
    bwidth, bheight = bbox.width, bbox.height

    inset_x0 = bx0 + bwidth*x0
    inset_y0 = by0 + bheight*y0
    inset_w = width * bwidth
    inset_h = height * bheight if height else inset_w

    ax = fig.add_axes([inset_x0, inset_y0, inset_w, inset_h])

    if ticksize:
        ax.tick_params(axis='both', which='major', labelsize=ticksize)
        ax.tick_params(axis='both', which='minor', labelsize=ticksize)

    return ax


def inset_adjacent_axes(base_ax, width, dw=.05, fig=None):

    if fig is None:
        fig = plt.gcf()

    bbox = base_ax.get_position()
    bx0, by0, bx1, by1 = bbox.x0, bbox.y0, bbox.x1, bbox.y1
    bwidth, bheight = bbox.width, bbox.height

    inset_x0 = bx1 + dw
    inset_x1 = bx1 + dw + width
    inset_y0 = by0
    inset_y1 = by1

    ax = fig.add_axes([inset_x0, inset_y0, inset_x1-inset_x0, inset_y1-inset_y0])

    return ax


def make_room(factor, side='top', ax=None):
    assert factor > 0

    if ax is None:
        ax = plt.gca()

    if side in ['top', 'bottom']:
        ylim = ax.get_ylim()

        if side == 'top':
            ylim_new = ylim[0], ylim[1] + factor*(ylim[1]-ylim[0])
        else:
            ylim_new = ylim[0] - factor*(ylim[1]-ylim[0]), ylim[1]

        ax.set_ylim(ylim_new)

        return ylim_new

    elif side in ['right', 'left']:
        xlim = ax.get_xlim()

        if side == 'right':
            xlim_new = xlim[0], xlim[1] + factor*(xlim[1]-xlim[0])
        else:
            xlim_new = xlim[0] - factor*(xlim[1]-xlim[0]), xlim[1]

        ax.set_xlim(xlim_new)

        return xlim_new

    else:
        raise ValueError('bad value for side: must be one of \'top\', \'bottom\', \'left\', \'right\'')


def generate_ticks(vmin, vmax, nbins=5):
    ticks = MaxNLocator(nbins=5)._raw_ticks(vmin, vmax)

    if ticks[-1] > vmax:
        ticks = ticks[:-1]

    return ticks


def segment_plot(x_, y_, yval=1, ax=None, label=None, **kwargs):
    # TODO: augment with make_room and place on top of plot automatically

    x_ = np.asarray(x_)
    y_ = np.asarray(y_, dtype=int)

    assert x_.ndim == 1
    assert y_.ndim == 1
    assert x_.shape == y_.shape

    assert y_.max() <= 1
    assert y_.min() >= 0

    d_ = np.pad(np.diff(y_), (1, 0), mode='constant')

    start_i = np.where(d_ == 1)[0]
    if y_[0] == 1:
        start_i = np.pad(start_i, (1, 0), mode='constant')

    end_i = np.where(d_ == -1)[0]
    if y_[-1] == 1:
        end_i = np.pad(end_i, (0, 1), mode='constant', constant_values=len(x_))
    end_i -= 1

    start_x = x_[start_i]
    end_x = x_[end_i]

    assert len(start_x) == len(end_x)

    if ax is None:
        ax = plt.gca()

    for k, (x0, x1) in enumerate(zip(start_x, end_x)):
        ax.plot([x0, x1], [yval, yval], label=label if k == 0 else None, **kwargs)


def is_visible(obj, require_full_visibility=False, ax=None):
    if ax is None:
        ax = plt.gca()

    x, y = obj.get_data()

    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()

    if require_full_visibility:
        x_vis = (x.max() > xmin and x.min() < xmax)
        y_vis = (y.max() > ymin and y.min() < ymax)
    else:
        x_vis = (x.min() > xmin and x.max() < xmax)
        y_vis = (y.min() > ymin and y.max() < ymax)

    return (x_vis and y_vis)


# TODO: buggy, fix or remove: not needed badly because one can get the color of
# the lines after using plt.plot()
def get_next_style(ax=None):
    if ax is None:
        ax = plt.gca()

    cc = ax._get_lines.prop_cycler
    n = next(cc)

    c = n['color']
    ls = n['linestyle'] if 'linestyle' in n.keys() else None

    return c, ls


def plot_without_autoscale(x, y, ax=None, c=None, ls=None, **kwargs):
    # if c is None, get style from prop cycle of axes
    # note: can be very slow in some cases

    if ax is None:
        ax = plt.gca()

    x = np.asarray(x)
    y = np.asarray(y)
    assert x.ndim == 1
    assert y.ndim == 1
    assert x.shape == y.shape

    assert c is None or 'color' not in kwargs.keys(), 'can specify c or color argument, not both'
    assert ls is None or 'linestyle' not in kwargs.keys(), 'can specify ls or linestyle argument, not both'

    if 'color' in kwargs.keys():
        # use color as specified by keyword color
        pass
    else:
        if c is None:
            # get next style from color cycle
            c, ls_from_cc = get_next_style(ax)
            if ls_from_cc and ls is None:  # don't override
                ls = ls_from_cc

        kwargs['color'] = c
        if ls:
            kwargs['linestyle'] = ls

    xy = np.column_stack((x, y))
    ax.add_collection(LineCollection([xy], **kwargs), autolim=False)


def get_tiled_plot(num_tiles, tile_size, ravel=True, **kwargs):
    r = int(np.ceil(np.sqrt(num_tiles)))
    tiles = (r, r-1) if r*(r-1) >= num_tiles else (r, r)
    assert np.prod(tiles) >= num_tiles

    figsize = (tiles[0]*tile_size[0], tiles[1]*tile_size[1])

    fig, ax_ = plt.subplots(*tiles, figsize=figsize, **kwargs)

    if ravel:
        ax_ = ax_.ravel()

    return fig, ax_


def membrane_potential_with_spikes(t_, V_, spikes, V_spike=-40):
    V_plot_ = V_.copy()

    for k, t in enumerate(t_):
        if t in spikes:
            V_plot_[k] = -40

    return V_plot_
